"""
gidown.search
~~~~~~~~~~~~~
This module implements functions and classes for acquiring images from google image search.

:license: Apache2, see LICENSE for more details.
"""

import json
from imghdr import what
from typing import List
from urllib.parse import quote_plus

import requests
from bs4 import BeautifulSoup

from gidown.query import QueryBuilder
from gidown.advanced import QuerySettings, QuerySetting


class GoogleSearchImage:

    """
    Class used to store information about images from google image search. 
    Two **GoogleSearchImage** are considered the same if they have the same image URL.
    """

    def __init__(self, google_json_dict: dict):
        """
        
        
        :param google_json_dict: Dictionary as found in <div> for each image on google image search pages  
        """

        self.url = google_json_dict["ou"]
        self.tb_url = google_json_dict["tu"]

        self.src_url = google_json_dict["ru"]
        self.src_domain = google_json_dict["isu"]

        self.title = "pt"
        self.desc = "s"

        self.width = google_json_dict["ow"]
        self.height = google_json_dict["oh"]
        self.type = google_json_dict["ity"]

        self.tb_width = google_json_dict["tw"]
        self.tb_height = google_json_dict["th"]
        self.tb_type = google_json_dict["ity"]

        self.image = None
        self.thumbnail = None

    def __eq__(self, other):
        if not hasattr(other, "image_url"):
            return False
        return self.url == other.image_url

    def __hash__(self):
        return hash(self.url)

    def __str__(self):
        return "Downloadable {} at {}".format(self.type, self.url)

    def download(self, download_all=False, check_ext=True) -> None:
        """
        Download the image and store raw bytes into **self.image**.
        Image type (extension) is automatically detected from the raw bytes if possible (**type**).

        :param check_ext: Checks if extensions is correct. Raises ValueError if extension cannot be determined from image.
        :param download_all: Also download thumbnail and store it in **self.thumbnail**.
        """

        self.image = requests.get(self.url).content
        if download_all:
            self.download_thumbnail()
        ext = what(None, self.image)
        if check_ext and ext is None:
            raise ValueError("Unknown file format.")
        self.type = ext if ext is not None else self.type

    def download_thumbnail(self) -> None:
        """
        Download the thumbnail and store raw bytes into **self.thumbnail**.
        Image type (extension) is automatically detected from the raw bytes if possible (**thumbnail_type**).
        """
        self.thumbnail = requests.get(self.tb_url).content
        ext = what(None, self.image)
        self.tb_type = ext if ext is not None else self.tb_type

    @staticmethod
    def _save(img_data, filename):
        with open(filename, "wb") as fout:
            fout.write(img_data)

    def save(self, filename: str, auto_ext=False):
        """
        Save image to filesystem. If the image was not downloaded using **download** it is done now.
               
        Image type (extension) is automatically detected from the raw bytes if possible, otherwise the type is
        determined from the google image search dictionary.

        :param filename: Path to store the image to.
        :param auto_ext: Automatically add the file extension.
        """
        if self.image is None:
            self.download()

        if auto_ext:
            filename = "{}.{}".format(filename, self.type)
        self._save(self.image, filename)

    def save_thumbnail(self, filename, auto_ext=False):
        """
        Save thumbnail to filesystem. If the thumbnail was not downloaded using **download_thumbnail** it is done now.
               
        Image type (extension) is automatically detected from the raw bytes if possible, otherwise the type is
        determined from the google image search dictionary.

        :param filename: Path to store the thumbnail to.
        :param auto_ext: Automatically add the file extension.
        """
        if self.thumbnail is None:
            self.download_thumbnail()

        if auto_ext:
            filename = "{}.{}".format(filename, self.tb_type)
        self._save(self.thumbnail, filename)


_USER_AGENT = "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.36"
_URL = "https://www.google.hr/search"


def _generate_url(query, settings, autocorrect=False):
    q = {"tbm": "isch",
         "tbs": QuerySettings(settings).urlencode(),
         "q": quote_plus(query.strip()),
         "source": "lnms"}
    if not autocorrect:
        q["nfpr"] = 1
    return "{}?{}".format(_URL, "&".join("{}={}".format(k, v) for k, v in q.items()))


def _fetch_html(url):
    req = requests.get(url, headers={'User-Agent': _USER_AGENT})
    html_doc = req.text
    return html_doc


def image_query(query: str or QueryBuilder, *restrictions: QuerySetting, autocorrect=False) -> List[GoogleSearchImage]:
    """
    Query google for images that satisfy all the restrictions.

    :param query: Query string, supports all advanced google search options.(see gis.query for wrappers around often used options) 
    :param restrictions: Variable amount of advanced search options. (see gis.advanced)
    :param autocorrect: Auto-correct queries that have a high likelihood of a misspell (e.q. "color grenn")
    :return: List of unique GoogleSearchImage objects that can be downloaded and saved.
    """
    url = _generate_url(query, restrictions, autocorrect)
    html = _fetch_html(url)

    soup = BeautifulSoup(html, 'html.parser')
    divs = soup.find_all("div", {"class": "rg_meta"})

    data = [json.loads(div.text) for div in divs]
    return list(set(GoogleSearchImage(img_data) for img_data in data))



"""
gidown.advanced
~~~~~~~~~~~~~~~
This module contains advanced image search options.

:license: Apache2, see LICENSE for more details.
"""

import datetime
import re


class QuerySetting:

    """
    General class to represent a single advanced search option. 
    Generally shouldn't be used outside the scope of this file.
    """

    def __init__(self, setting_str: str):
        """       
        General class to represent a single advanced search option. 
        
        :param setting_str: String as it appears in tbs argument of google image search URL. Format: "key":"value"
        
        """

        if setting_str and not re.match("^.+:.+$", setting_str):
            raise ValueError("Setting string must be in the format <key>:<value>, {} was given".format(setting_str))
        self._setting = setting_str

    def __str__(self):
        return self._setting

    def __repr__(self):
        return str(self)

    def urlencode(self) -> str:
        """
        Encode setting to match the format in tbs argument of google image search URL.
        Simply returns the setting_str as it was given.
        
        :return: Setting encoded to match the format in tbs argument of google image search URL
        """
        return self._setting

    def __add__(self, other):
        if isinstance(other, QuerySettings):
            return other + self
        if isinstance(other, self.__class__):
            return QuerySettings(self, other)


class QuerySettings:

    """
    General class to represent multiple advanced search options.
    """

    def __init__(self, *settings: str or QuerySetting or "QuerySettings"):
        """
        Class to represent multiple advanced search options.
        
        :param settings: Variable amount of QuerySetting, QuerySettings objects or strings.  
        """
        self._settings = list(*settings)

    def __str__(self):
        return " and ".join(str(setting) for setting in self._settings)

    def __repr__(self):
        return str(self)

    def urlencode(self) -> str:
        """ 
        Encode settings to match the format in **tbs** argument of google image search URL.

        :return: Setting encoded to match the format in tbs argument of google image search URL
        """
        return ",".join(setting.urlencode() for setting in self._settings)

    def __add__(self, other):
        if isinstance(other, self.__class__):
            ret = QuerySettings()
            ret._settings = self._settings + other._settings
            return ret
        if isinstance(other, QuerySetting):
            ret = QuerySettings()
            ret._settings = self._settings[:]
            ret._settings.append(other)
            return ret


class Size:

    """
    Collection of **image size** restrictions that can be used in advanced search options. 
    """

    def __init__(self):
        """This class was not meant to be instanced."""
        raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

    ANY = QuerySetting("")
    """Any size, doesn't effect search results."""
    LARGE = QuerySetting("isz:l")
    """Restrict search to large images"""
    MEDIUM = QuerySetting("isz:m")
    """Restrict search to medium images."""
    ICON = QuerySetting("isz:i")
    """Restrict search to icons."""

    @staticmethod
    def exactly(w: int, h: int) -> QuerySetting:
        """
        Restrict search to images of exact size.

        :param w: Required image width
        :param h: Required image height
        :return: QuerySetting representing desired restriction
        """
        return QuerySetting("isz:ex,iszw:{},iszh:{}".format(w, h))

    class LargerThen:

        """
        Collection of **larger-then image size** restrictions.
        """

        def __init__(self):
            """This subclass was not meant to be instanced."""
            raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

        S_400x300 = QuerySetting("isz:lt,islt:qsvga")
        """Restrict search to QSVGA (400x300) images."""
        S_640x480 = QuerySetting("isz:lt,islt:vga")
        """Restrict search to VGA (640x480) images."""
        S_800x600 = QuerySetting("isz:lt,islt:svg")
        """Restrict search to SVG (800x600) images."""
        S_1024x768 = QuerySetting("isz:lt,islt:xga")
        """Restrict search to XGA (1024x768) images."""

        QSVGA = S_400x300
        """Restrict search to QSVGA (400x300) images."""
        VGA = S_640x480
        """Restrict search to VGA (640x480) images."""
        SVG = S_800x600
        """Restrict search to SVG (800x600) images."""
        XGA = S_1024x768
        """Restrict search to XGA (1024x768) images."""

        MP_2 = QuerySetting("isz:lt,islt:2mp")
        """Restrict search to 2MP images."""
        MP_4 = QuerySetting("isz:lt,islt:4mp")
        """Restrict search to 4MP images."""
        MP_6 = QuerySetting("isz:lt,islt:6mp")
        """Restrict search to 6MP images."""
        MP_8 = QuerySetting("isz:lt,islt:8mp")
        """Restrict search to 8MP images."""
        MP_10 = QuerySetting("isz:lt,islt:10mp")
        """Restrict search to 10MP images."""
        MP_12 = QuerySetting("isz:lt,islt:12mp")
        """Restrict search to 12MP images."""
        MP_15 = QuerySetting("isz:lt,islt:15mp")
        """Restrict search to 15MP images."""
        MP_20 = QuerySetting("isz:lt,islt:20mp")
        """Restrict search to 20MP images."""
        MP_40 = QuerySetting("isz:lt,islt:40mp")
        """Restrict search to 40MP images."""
        MP_70 = QuerySetting("isz:lt,islt:70mp")
        """Restrict search to 70MP images."""


class Color:

    """
    Collection of **dominant image color** restrictions that can be used in advanced search options. 
    """

    def __init__(self):
        raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

    ANY = QuerySetting("")
    """No color restrictions, doesn't effect search results."""
    FULL = QuerySetting("ic:color")
    """Restrict search to full color images."""
    GRAY_SCALE = QuerySetting("ic:gray")
    """Restrict search to black and white images."""
    TRANSPARENT = QuerySetting("ic:trans")
    """Restrict search to partially transparent images."""

    RED = QuerySetting("ic:specific,isc:red")
    """Restrict search to dominantly red images."""
    ORANGE = QuerySetting("ic:specific,isc:orange")
    """Restrict search to dominantly orange images."""
    YELLOW = QuerySetting("ic:specific,isc:yellow")
    """Restrict search to dominantly yellow images."""
    GREEN = QuerySetting("ic:specific,isc:green")
    """Restrict search to dominantly green images."""
    TEAL = QuerySetting("ic:specific,isc:teal")
    """Restrict search to dominantly teal images."""
    BLUE = QuerySetting("ic:specific,isc:blue")
    """Restrict search to dominantly blue images."""
    PURPLE = QuerySetting("ic:specific,isc:purple")
    """Restrict search to dominantly purple images."""
    PINK = QuerySetting("ic:specific,isc:pink")
    """Restrict search to dominantly pink images."""
    WHITE = QuerySetting("ic:specific,isc:white")
    """Restrict search to dominantly white images."""
    GRAY = QuerySetting("ic:specific,isc:gray")
    """Restrict search to dominantly gray images."""
    BLACK = QuerySetting("ic:specific,isc:black")
    """Restrict search to dominantly black images."""
    BROWN = QuerySetting("ic:specific,isc:brown")
    """Restrict search to dominantly brown images."""


class Type:

    """
    Collection of **image type** restrictions that can be used in advanced search options. 
    """

    def __init__(self):
        raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

    ANY = QuerySetting("")
    """No type restrictions, doesn't effect search results."""
    FACE = QuerySetting("itp:face")
    """Restrict search to images of faces."""
    PHOTO = QuerySetting("itp:photo")
    """Restrict search to photos."""
    CLIP_ART = QuerySetting("itp:clipart")
    """Restrict search to clipart."""
    LINE_DRAWING = QuerySetting("itp:lineart")
    """Restrict search to lineart."""


class Time:

    """
    Collection of **image upload time** restrictions that can be used in advanced search options. 
    """

    def __init__(self):
        raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

    ANY = QuerySetting("")
    """No time restrictions, doesn't effect search results."""
    PAST_SECOND = QuerySetting("qdr:d")
    """Restrict search to images that were uploaded in the past second."""
    PAST_24H = QuerySetting("qdr:d")
    """Restrict search to images that were uploaded in the past 24 hours."""
    PAST_WEEK = QuerySetting("qdr:w")
    """Restrict search to images that were uploaded in the past week."""

    _valid_measures = "nhdwm"

    @staticmethod
    def past(amount: int, unit="h"):
        """
        Restrict search to images that were uploaded in the past <amount> <unit>.
        
        To get results from the past 10 weeks you would add the restriction:
        >>> from gidown.advanced import Time
        >>> Time.past(10, "w")
        tbs=qdr:10w
        
        Allowed units of time:
        
        * n - minute
        * h - hour
        * d - day
        * w - week
        * m - month

        :param amount: Amount of time to look into the past, measured <unit>.
        :param unit: One of the allowed units of time. Defaults to hour (h).
        :return: QuerySetting representing desired restriction.
        """
        if unit not in Time._valid_measures:
            raise ValueError("Unknown measure {}, must be one of: {}".format(unit, Time._valid_measures))
        return QuerySetting("tbs=qdr:{}{}".format(unit, "" if amount <= 1 else amount))

    @staticmethod
    def range(date_from: datetime.date, date_to: datetime.date) ->QuerySetting:
        """
        Generates generic time restrictions. Include images uploaded from <date_from> to <date_to>.

        :param date_from: Starting date, inclusive.
        :param date_to: Ending date, inclusive.
        :return: QuerySetting representing desired restriction.
        """
        return QuerySetting("cdr=1,cd_min={}.{}.{},cd_max={}.{}.{}".format(
            date_from.day, date_from.month, date_from.year, date_to.day, date_to.month, date_to.year))


class FileFormat:

    """
    Collection of **file format** restrictions that can be used in advanced search options.
    """

    def __init__(self):
        raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

    ANY = QuerySetting("")
    """No format restrictions, doesn't effect search results."""
    JPG = QuerySetting("ift:jpg")
    """Restrict search to JPG files."""
    GIF = QuerySetting("ift:gif")
    """Restrict search to GIF files."""
    PNG = QuerySetting("ift:png")
    """Restrict search to PNG files."""
    BMP = QuerySetting("ift:mbp")
    """Restrict search to BMP files."""
    SVG = QuerySetting("ift:svg")
    """Restrict search to SVG files."""
    WEBP = QuerySetting("ift:webp")
    """Restrict search to WEBP files."""
    ICO = QuerySetting("ift:ico")
    """Restrict search to ICO files."""


class AspectRatio:

    """
    Collection of **aspect ratio** restrictions that can be used in advanced search options.
    """

    def __init__(self):
        raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

    ANY = QuerySetting("")
    """No aspect ratio restrictions, doesn't effect search results."""
    TALL = QuerySetting("iar:t")
    """Restrict search to tall images. (width < height)"""
    SQUARE = QuerySetting("iar:s")
    """Restrict search to square images. (width == height)"""
    WIDE = QuerySetting("iar:w")
    """Restrict search to wide images. (width > height)"""
    PANORAMIC = QuerySetting("iar:xw")
    """Restrict search to panoramic images. (width >> height)"""


class UsageRights:

    """
    Collection of **Usage rights** restrictions that can be used in advanced search options.
    """

    def __init__(self):
        raise NotImplementedError()

    ANY = QuerySetting("")
    """No usage rights restriction, doesn't effect search results."""
    REUSE_WITH_MODIFICATION = QuerySetting("sur:fc")
    """Restrict search to images that can be used commercially with modification"""
    REUSE = QuerySetting("sur:fmc")
    """Restrict search to images that can be used commercially without modification"""
    NONCOMMERCIAL_REUSE_WITH_MODIFICATION = QuerySetting("sur:fm")
    """Restrict search to images that can be used non-commercially with modification"""
    NONCOMMERCIAL_REUSE = QuerySetting("sur:f")
    """Restrict search to images that can be used non-commercially without modification"""


class Sorting:

    """
    Collection of **sorting** options that can be used in advanced search options.
    """

    def __init__(self):
        raise NotImplementedError("The {} class is not meant to be instanced.".format(self.__class__.__name__))

    BY_DATE = QuerySetting("sbd:1")
    """Sort results by date."""
    BY_RELEVANCE = QuerySetting("sbd:0")
    """Sort results by relevance."""



"""
Module that implements a simple interface for accessing Google Image Search.

File extensions are automatically determined from the image data or, if that fails,
directly from the results of the image search, as Google wa kind enough to provide a JSON with all information
about the image (image URL, source page URL, Type, Width, Height, Thumbnail URL, ...).

"""

from gidown.search import *
from gidown.query import QueryBuilder



"""
gidown.query
~~~~~~~~~~~~
This module implements advances query options such as site restriction and excluding words.

:license: Apache2, see LICENSE for more details.
"""
from io import StringIO


def all_of(*words: str) -> str:
    """
    Format words to query results containing all the desired words.

    :param words: List of desired words
    :return: String in the format google understands
    """
    return " ".join(*words)


def phrase(term: str) -> str:
    """
    Format words to query results containing the desired phrase.
    
    :param term: A phrase that appears in that exact form. (e.q. phrase "Chicago Bulls" vs. words "Chicago" "Bulls")
    :return: String in the format google understands
    """
    return '"{}"'.format(term)


def any_of(*words: str) -> str:
    """
    Format words to query results containing any the desired words.
    
    :param words: List of desired words
    :return: String in the format google understands
    """
    return " OR ".join(*words)


def none_of(*words: str) -> str:
    """
    Format words to query results containing none of the undesired words.
    This also works with the **on_site** restriction to exclude undesired domains.

    :param words: List of undesired words
    :return: String in the format google understands
    """
    return " ".join("-{}".format(word) for word in words)


def on_site(domain: str) -> str:
    """
    Restrict search to a specific domain.
    Can be repeated to restrict search on multiple sites.

    :param domain: Domain name
    :return: String in the format google understands
    """
    return "site:{}".format(domain)


class QueryBuilder:
    """
    Class for building queries similar to StringBuilder in Java. 
    
    Calling any function defined in :class:`str` on this class returns a string.

    >>> from gidown.query import QueryBuilder
    >>> q = QueryBuilder("text    ")
    >>> q.strip()
    'text'
    
    The query: *dog OR cat eats "ice cream" site:example.com* can be generated with:
    >>> from gidown.query import all_of, any_of, phrase, on_site
    >>> all_of(any_of("dog", "cat"), "eats", phrase("ice cream"), on_site("example.com"))
    
    or with:
    >>> from gidown.query import QueryBuilder
    >>> QueryBuilder().add_choice("dog", "cat").add("eats").add_phrase("ice cream").add_site_restriction("example.com")
       
    """

    def __init__(self, initial_query=""):
        self._query = StringIO()
        print(initial_query.strip(), sep=" ", end=" ", file=self._query)

    def add(self, *words) -> "QueryBuilder":
        """
        Add words to query.
        
        Wrapper around `gis.query.all_of`.

        :param words: Desired words
        :return: self
        """

        print(all_of(*words), sep=" ", end=" ", file=self._query)
        return self

    def add_phrase(self, term) -> "QueryBuilder":
        """
        Add a specific phrase to the query.
        
        Wrapper around `.gis.query.phrase`.

        :param term: Desired phrase
        :return: self
        """
        print(phrase(term), sep=" ", end=" ", file=self._query)
        return self

    def add_choice(self, *words) -> "QueryBuilder":
        """
        Add words to query as *any of*.

        Wrapper around gis.query.any_of.
        
        :param words: Desired words
        :return: self
        """
        print(any_of(*words), sep=" ", end=" ", file=self._query)
        return self

    def exclude(self, *words) -> "QueryBuilder":
        """
        Exclude results with undesired words.

        Wrapper around gis.query.none_of.

        :param words: 
        :return: self
        """
        print(none_of(*words), sep=" ", end=" ", file=self._query)
        return self

    def add_site_restriction(self, domain, exclude=False) -> "QueryBuilder":
        """
        Restrict search to a specific domain.
        Can be repeated to restrict search on multiple sites.

        Wrapper around gis.query.on_site.
        
        :param domain: Domain name that is searched.
        :param exclude: Exclude results from this site.
        :return: self
        """
        print(on_site(domain) if not exclude else none_of(on_site(domain)), sep=" ", end=" ", file=self._query)
        return self

    def __str__(self):
        return self._query.getvalue().strip()

    def __getattr__(self, name):
        if hasattr(str, name):
            func = getattr(str, name)
            return lambda *args, **kwargs: func(str(self), *args, **kwargs)
        raise AttributeError(name)



from setuptools import setup
import pip.req

try:
    requirements = pip.req.parse_requirements('requirements.txt')
except:
    requirements = pip.req.parse_requirements('requirements.txt', session=pip.download.PipSession())

setup(
    name="gidown",
    version='0.1.0',
    author='BonsAI d.o.o',
    author_email='opensource@bonsai.hr',
    description='Wrapper library for Google Image Search',
    license='Apache 2.0',
    keywords="google image search download wrapper",
    url='https://github.com/bonsaihr/gidown',
    packages=['gidown'],
    zip_safe=True,
    install_requires=[
        'beautifulsoup4',
        'requests'
      ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
    ],
)
